/*
 * Copyright (C) 2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "para.h"
#include <ortp/ortp.h>
#include "send.cmdline.h"

#define CHUNK_SIZE 65536

static RtpSession *session;
static char *header_buf;
static struct gengetopt_args_info conf;

enum {END_OF_FILE, BEGIN_OF_FILE, HEADER};

static void send_ctrlmsg(char *buf, int len, int ts)
{
	rtp_session_set_payload_type(session, PAYLOAD_OTHER);
	rtp_session_send_with_ts(session, buf, len, ts);
	rtp_session_set_payload_type(session, PAYLOAD_AUDIO_CONTINUOUS);
}

void para_log(int, char*, ...) __attribute__((format (printf, 2, 3)));
void para_log(int ll, char* fmt,...)
{
	va_list argp;

	/* ignore log message if loglevel is not high enough */
	if (ll < conf.loglevel_arg)
		return;
	va_start(argp, fmt);
	vfprintf(stderr, fmt, argp);
	va_end(argp);
}

static void send_notifier(int code, int ts)
{
	char buf[2];

	PARA_INFO_LOG("sending notifier #%d\n", code);
	if (code == HEADER) {
		*header_buf = HEADER;
		return send_ctrlmsg(header_buf, conf.header_length_arg + 1, ts);
	}
	buf[0] = code;
	buf[1] = '\0';
	send_ctrlmsg(buf, 2, ts);
}

static int save_header(int in, char *buf)
{
	int ret = -1, len = conf.header_length_arg;
	if (len > CHUNK_SIZE || len <= 0)
		return ret;
	free(header_buf);
	header_buf = malloc(len + 1);
	ret = read(in, buf, CHUNK_SIZE);
	if (ret < len) {
		PARA_ERROR_LOG("read returned %d/%d (%s)", ret, len,
			ret < 0? strerror(errno) : "success");
		return -2;
	}
	*header_buf = HEADER;
	memcpy(header_buf + 1, buf, len);
	return ret;
}

static int send_initial_buffer(char *buf, size_t len)
{
	int ts, ret;

	PARA_NOTICE_LOG("initial buffer: %d bytes\n", len);
	ts = rtp_session_get_current_send_ts(session);
	send_notifier(BEGIN_OF_FILE, ts);
	while (len > 0) {
		int send = MIN(len, 1000);
		ts = rtp_session_get_current_send_ts(session);
		ret = rtp_session_send_with_ts(session, buf, send, ts);
		PARA_NOTICE_LOG("sent %d/%d bytes with ts %d\n", send, ret, ts);
		usleep(1000 * 20);
		len -= send;
		memmove(buf, buf + send, len);
	}
	return 1;
}

static int do_io(int in, RtpSession *session)
{
	int ts = 0, sent = 0, ret, bytes, count = 0, old_ts = 0, header_ts = 0;
	char buf[CHUNK_SIZE];

	if (conf.header_length_given) {
		ret = save_header(in, buf);
		if (ret < 0) {
			PARA_ERROR_LOG("could not read header: %d\n", ret);
			return ret;
		}
		send_initial_buffer(buf, ret);
	}
	while (1) {
		ret = read(in, buf, CHUNK_SIZE);
		if (ret <= 0)
			break;
		count++;
		ts = rtp_session_get_current_send_ts(session);
		bytes = rtp_session_send_with_ts(session, buf, ret, ts);
		PARA_NOTICE_LOG("sent %d/%d bytes at ts %d\n", ret, bytes, ts);
		old_ts = ts;
		sent += ret;
		if (!conf.header_length_given)
			continue;
		if (ts > header_ts + conf.header_interval_arg * 5000) {
			PARA_INFO_LOG("sending headers (%d)\n",
				conf.header_length_arg);
			send_notifier(HEADER, ts);
			header_ts = ts;
		}
	}
	ts = rtp_session_get_current_send_ts(session);
	send_notifier(END_OF_FILE, ts);
	PARA_INFO_LOG("sent %d bytes\n", sent);
	return ret;
}

int main(int argc, char *argv[])
{
	int ret, infile;

	ret = cmdline_parser(argc, argv, &conf);
	if (conf.header_length_given) {
		PARA_NOTICE_LOG("assuming header len == %d\n",
			conf.header_length_arg);
	}
	PARA_NOTICE_LOG("sending to udp %s:%d\n",
		 conf.hostname_arg,conf.port_arg);
	ortp_init();
	ortp_scheduler_init();
	session = rtp_session_new(RTP_SESSION_SENDONLY);
	rtp_session_set_scheduling_mode(session, 1);
	rtp_session_set_blocking_mode(session, 1);
	rtp_session_set_remote_addr(session, conf.hostname_arg,conf.port_arg);
	rtp_session_set_payload_type(session, PAYLOAD_AUDIO_CONTINUOUS);

	if (conf.inputs_num) {
		infile = open(conf.inputs[0], O_RDONLY);
		ret = -1;
		if (!infile) {
			PARA_ERROR_LOG("%s", "can not open file\n");
			goto out;
		}
	} else
		infile = STDIN_FILENO;
	ret = do_io(infile, session);
	if (ret < 0)
		goto out;
	if (infile != STDIN_FILENO)
		close(infile);
out:
	rtp_session_destroy(session);
	ortp_exit();
	return ret;
}
